/*-
 * Copyright (c) 2011 Andrey Zonov <andrey@zonov.org>
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#define _RLIMIT_IDENT 1

#include <sys/mman.h>
#include <sys/param.h>
#include <sys/proc.h>
#include <sys/sysctl.h>
#include <sys/user.h>
#include <sys/types.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <sys/resourcevar.h>

#include <err.h>
#include <fcntl.h>
#include <kvm.h>
#include <paths.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

struct plimit {
	struct	rlimit pl_rlimit[RLIM_NLIMITS];
	int	pl_refcnts;
};

int
main(int argc, char **argv)
{
	int i;
	int fd;
	int cnt;
	off_t off;
	kvm_t *kd;
	rlim_t lim;
	char *endptr;
	struct kinfo_proc *kip;
	struct proc *p;
	struct plimit *plimits;

	if (argc < 3 || argc > 4)
		errx(1, "usage: chlim <pid> <ident> [val]");

	kd = kvm_open(NULL, _PATH_MEM, NULL, O_RDONLY, NULL);
	if (kd == NULL)
		err(1, "kvm_open()");

	kip = kvm_getprocs(kd, KERN_PROC_PID, atoi(argv[1]), &cnt);
	if (kip == NULL || cnt != 1)
		err(1, "kvm_getprocs()");

	fd = open(_PATH_KMEM, O_RDWR, 0);
	if (fd == -1)
		err(1, "open()");

	off = (off_t)((uintptr_t)kip->ki_paddr);
	p = mmap(0, sizeof(struct proc), PROT_READ | PROT_WRITE, MAP_SHARED, fd, off);
	if (p == MAP_FAILED)
		err(1, "mmap()");

	off = (off_t)((uintptr_t)p->p_limit);
	plimits = mmap(0, sizeof(struct plimit), PROT_READ | PROT_WRITE, MAP_SHARED, fd, off);
	if (plimits == MAP_FAILED)
		err(1, "mmap()");

	for (i = 0; i < RLIM_NLIMITS; i++) {
		if (strcmp(argv[2], rlimit_ident[i]) == 0) {
			if (argc == 4) {
				lim = strtol(argv[3], &endptr, 10);
				if (lim == -1)
					lim = RLIM_INFINITY;
				if (lim < -1) {
					printf("wrong limit value, use -1..%ld\n", RLIM_INFINITY);
					exit(1);
				}
				plimits->pl_rlimit[i].rlim_cur = lim;
				plimits->pl_rlimit[i].rlim_max = lim;
				printf("%s soft and hard limits are set: %ld\n", rlimit_ident[i],
				    lim == RLIM_INFINITY ? -1 : lim);
			} else {
				lim = plimits->pl_rlimit[i].rlim_cur;
				printf("%s %ld\n", rlimit_ident[i], lim == RLIM_INFINITY ? -1 : lim);
			}
			exit(0);
		}
	}

	printf("wrong ident, use one of this: ");
	for (i = 0; i < RLIM_NLIMITS; i++)
		printf("%s ", rlimit_ident[i]);
	printf("\n");

	exit(1);
}
